@@ -4,8 +4,10 @@
 from __future__ import absolute_import, division, print_function
 import urwid
 import bdb
-import sys
+import gc
 import os
+import sys
+from types import TracebackType
 
 from pudb.settings import load_config, save_config
 CONFIG = load_config()
@@ -192,7 +194,7 @@ class Debugger(bdb.Bdb):
                     break
                 frame = frame.f_back
 
-    def set_trace(self, frame=None, as_breakpoint=True):
+    def set_trace(self, frame=None, as_breakpoint=None, paused=True):
         """Start debugging from `frame`.
 
         If frame is not specified, debugging starts from caller's frame.
@@ -200,7 +202,19 @@ class Debugger(bdb.Bdb):
         Unlike Bdb.set_trace(), this does not call self.reset(), which causes
         the debugger to enter bdb source code. This also implements treating
         set_trace() calls as breakpoints in the PuDB UI.
+
+        If as_breakpoint=True (the default), this call will be treated like a
+        breakpoint in the UI (you can press 'b' on it to disable breaking
+        here).
+
+        If paused=False, the debugger will not break here.
         """
+        if as_breakpoint is None:
+            if not paused:
+                as_breakpoint = False
+            else:
+                as_breakpoint = True
+
         if frame is None:
             frame = thisframe = sys._getframe().f_back
         else:
@@ -223,7 +237,10 @@ class Debugger(bdb.Bdb):
                     self.ui.set_source_code_provider(
                             self.ui.source_code_provider, force_update=True)
 
-            self.set_step()
+            if paused:
+                self.set_step()
+            else:
+                self.set_continue()
             sys.settrace(self.trace_dispatch)
         else:
             return
@@ -309,6 +326,14 @@ class Debugger(bdb.Bdb):
     def interaction(self, frame, exc_tuple=None, show_exc_dialog=True):
         if exc_tuple is None:
             tb = None
+        elif isinstance(exc_tuple, TracebackType):
+            # For API compatibility with other debuggers, the second variable
+            # can be a traceback object.  In that case, we need to retrieve the
+            # corresponding exception tuple.
+            tb = exc_tuple
+            exc, = (exc for exc in gc.get_referrers(tb)
+                    if getattr(exc, "__traceback__", None) is tb)
+            exc_tuple = type(exc), exc, tb
         else:
             tb = exc_tuple[2]
 
@@ -512,7 +537,13 @@ class NullSourceCodeProvider(SourceCodeProvider):
                 SourceLine(debugger_ui, ""),
                 SourceLine(debugger_ui, "If this is generated code and you would "
                     "like the source code to show up here,"),
-                SourceLine(debugger_ui, "simply set the attribute "
+                SourceLine(debugger_ui, "add it to linecache.cache, like"),
+                SourceLine(debugger_ui, ""),
+                SourceLine(debugger_ui, "    import linecache"),
+                SourceLine(debugger_ui, "    linecache.cache[filename] = "
+                    "(size, mtime, lines, fullname)"),
+                SourceLine(debugger_ui, ""),
+                SourceLine(debugger_ui, "You can also set the attribute "
                     "_MODULE_SOURCE_CODE in the module in which this function"),
                 SourceLine(debugger_ui, "was compiled to a string containing "
                     "the code."),
@@ -1544,8 +1575,11 @@ class DebuggerUI(FrameVarInfoKeeper):
             sys.stdin = None
             sys.stderr = sys.stdout = StringIO()
             try:
+                # Don't use cmdline_get_namespace() here, it breaks things in
+                # Python 2 (issue #166).
                 eval(compile(cmd, "<pudb command line>", 'single'),
-                        cmdline_get_namespace())
+                        self.debugger.curframe.f_globals,
+                        self.debugger.curframe.f_locals)
             except:
                 tp, val, tb = sys.exc_info()
 
@@ -2095,7 +2129,7 @@ class DebuggerUI(FrameVarInfoKeeper):
                 self.message("Package 'pygments' not found. "
                         "Syntax highlighting disabled.")
 
-        WELCOME_LEVEL = "e030"  # noqa
+        WELCOME_LEVEL = "e031"  # noqa
         if CONFIG["seen_welcome"] < WELCOME_LEVEL:
             CONFIG["seen_welcome"] = WELCOME_LEVEL
             from pudb import VERSION
@@ -2112,6 +2146,9 @@ class DebuggerUI(FrameVarInfoKeeper):
                     "(invoked by hitting '?' after this message) should get you "
                     "on your way.\n"
 
+                    "\nChanges in version 2017.1.2:\n\n"
+                    "- Bug fixes.\n"
+
                     "\nChanges in version 2017.1.1:\n\n"
                     "- IMPORTANT: 2017.1 and possibly earlier versions had a \n"
                     "  bug with exponential growth of shell history for the \n"
